from polly.session import PollySession
from requests.adapters import HTTPAdapter, Retry
from polly import helpers
from polly import constants as const
from polly.help import example, doc

link_doc = "https://docs.elucidata.io/OmixAtlas/Polly%20Python.html"
retries = Retry(total=5, backoff_factor=0.1, status_forcelist=[500, 502, 503, 504])


class UnauthorizedException(Exception):
    """

    :meta private:
    """

    def __str__(self):
        return f"Authorization failed as credentials not found. Please use Polly.auth(token) as shown here  ---- {link_doc}"


class Polly:
    """
    This class for authorization to use Polly on any local or cloud platform. To authenticate usage of Polly, \
the following function can be used.
    """

    default_session = None
    example = classmethod(example)
    doc = classmethod(doc)

    @classmethod
    def auth(cls, token, env="", default_env="polly"):
        """
        Function for authorize usage of Polly on terminal or notebook.

        ``Args:``
            |  ``token (str):`` token copy from Polly front-end.


        ``Returns:``
            |  If token is not correct, it will give an error, else it will clear the authentication for user to\
 get started with polly-python.

        ``Error:``
            |  ``UnauthorizedException:`` when the token is expired.

        To use auth function import class Polly.

        .. code::


                from polly.auth import Polly
                Polly.auth(token)
        """
        # check if COMPUTE_ENV_VARIABLE present or not
        # if COMPUTE_ENV_VARIABLE, give priority
        env = helpers.get_platform_value_from_env(
            const.COMPUTE_ENV_VARIABLE, default_env, env
        )
        cls.default_session = PollySession(token, env=env)
        cls.default_session.mount(
            "https://",
            HTTPAdapter(pool_connections=100, pool_maxsize=100, max_retries=retries),
        )

    @classmethod
    def get_session(cls, token=None, env="polly"):
        """
        Function to get session from polly.

        ``Args:``
            |  ``token (str):`` token copy from polly.

        ``Returns:``
            |  if token is not satisfied it will throw UnauthorizedException.
            |  else it will return a polly.session object.

        ``Error:``
            |  ``UnauthorizedException:`` when the token is expired.

        To use get_sesion function import class Polly.


        .. code::


                from polly.auth import Polly
                session = Polly.get_session(token)

        """
        if not token:
            if not cls.default_session:
                raise UnauthorizedException
            else:
                return cls.default_session
        else:
            cls.auth(token, env=env)
            return cls.default_session



from polly import helpers
from polly.errors import (
    InvalidCohortOperationException,
    InvalidParameterException,
    InvalidRepoException,
    CohortEditException,
    EmptyCohortException,
    InvalidCohortAddition,
    InvalidCohortMergeOperation,
    InvalidPathException,
)
import os
import pandas as pd
from deprecated import deprecated
from polly.bridge_cohort import CohortRepoClass
from polly.core_cohort import CohortFileStandard
from polly import constants as const
from polly.help import example, doc


class Cohort:
    """
    The Cohort class contains functions which can be used to create cohorts, add or remove samples,\
merge metadata and data-matrix of samples/datasets in a cohort and edit or delete a cohort.

    ``Args:``
        |  ``token (str):`` token copy from polly.

    .. code::


            from polly.cohort import Cohort
            # To use this class initialize a object like this if you are not authorised on polly
            cohort = Cohort(token)

    If you are authorised then you can initialize object without token to know about :ref:`authentication <auth>`.
    """

    _cohort_info = None
    example = classmethod(example)
    doc = classmethod(doc)

    def __init__(self, token=None, env="", default_env="polly") -> None:
        # check if COMPUTE_ENV_VARIABLE present or not
        # if COMPUTE_ENV_VARIABLE, give priority
        env = helpers.get_platform_value_from_env(
            const.COMPUTE_ENV_VARIABLE, default_env, env
        )
        if self._cohort_info is None:
            self._cohort_info = helpers.get_cohort_constants()
        self._cohort_path = None
        self._cohort_details = None
        self.bridge_obj = CohortRepoClass(token, env)
        self.core_obj = CohortFileStandard()

    def create_cohort(
        self,
        local_path: str,
        cohort_name: str,
        description: str,
        repo_key=None,
        dataset_id=None,
        sample_id=None,
    ) -> None:
        """
            This function is used to create a cohort.

            ``Args:``
                |  ``local_path(str):`` local path to instantiate the cohort.
                |  ``cohort_name(str):`` identifier name for the cohort.
                |  ``description(str):`` description about the cohort.
                |  ``repo_key(str):`` Optional argument: repo_key(repo_name/repo_id) for the omixatlas from where \
datasets or samples is to be added.
                |  ``entity_id(list):`` Optional argument: list of dataset_id or sample_id to be added to the \
cohort.

            ``Returns:``
                |  A confirmation message on creation of cohort.

            ``Errors:``
                |  ``InvalidParameterException:`` Empty or Invalid Parameters
                |  ``InvalidCohortNameException:`` The cohort_name does not represent a valid cohort name.
                |  ``InvalidPathException:`` Provided path does not represent a file or a directory.

            | After making Cohort Object you can create cohort.
            | Example-
            | 1. while passing argument of repo and dataset.

            .. code::


                cohort.create_cohort("/import/tcga_cohort","cohort_name","cohort_description",\
"repo_id",list_of_datasets)

            | 2. without passing argument of repo and dataset.

            .. code::


                    cohort2.create_cohort("/path","cohort_name","cohort_description")


        """
        self._cohort_path = self.bridge_obj.create_cohort(
            local_path, cohort_name, description
        )
        self._cohort_details = self.bridge_obj._cohort_details
        if repo_key:
            self.add_to_cohort(
                repo_key=repo_key,
                dataset_id=dataset_id,
                sample_id=sample_id,
            )

    def add_to_cohort(self, repo_key: str, dataset_id=None, sample_id=None) -> None:
        """
        This function is used to add datasets or samples to a cohort.

        ``Args:``
            |  ``repo_key(str):`` repo_key(repo_name OR repo_id) for the omixatlas where datasets or samples belong.
            |  ``entity_id(list):`` list of dataset ID or sample ID to be added to the cohort.

        ``Returns:``
            |  A confirmation message for number of datasets or samples which are added to the cohort.

        ``Errors:``
            |  ``InvalidParameterException:`` Empty or Invalid Parameters.
            |  ``InvalidCohortOperationException:`` This operation is not valid as no cohort has been \
instantiated.

        | After creating cohort we can add datasets or samples to cohort.
        | Example-
        .. code::


                cohort.add_to_cohort("repo_id",list_of_dataset_ids)

        """
        self.bridge_obj._cohort_details = self._cohort_details
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        if not ((repo_key and dataset_id and sample_id) or (repo_key and dataset_id)):
            raise InvalidParameterException("repo_key dataset_id sample_id")
        repo_name = self.bridge_obj._return_omixatlas_name(repo_key)
        if repo_name not in self._cohort_info:
            raise InvalidRepoException(repo_name)
        cohort_entity_type = self._assign_entity_type(repo_name, dataset_id)
        if "entity_type" in self._cohort_details:
            existing_entity_type = self._cohort_details.get("entity_type")
            if existing_entity_type != cohort_entity_type:
                raise InvalidCohortAddition
        self.bridge_obj._cohort_details = self._cohort_details
        if cohort_entity_type == "dataset":
            self.bridge_obj.add_to_cohort_single(
                self._cohort_path,
                repo_name=repo_name,
                entity_id=dataset_id,
                entity_type=cohort_entity_type,
            )
        else:
            self.bridge_obj.add_to_cohort_multiple(
                self._cohort_path,
                repo_name=repo_name,
                dataset_id=dataset_id,
                sample_id=sample_id,
                entity_type=cohort_entity_type,
            )
        self._cohort_details = self.bridge_obj._cohort_details

    def remove_from_cohort(self, dataset_id=None, sample_id=[]) -> None:
        """
        This function is used for removing datasets or samples from a cohort.

        ``Args:``
            |  ``entity_id(list):`` list of dataset IDs or sample IDs which is to be removed from the cohort.

        ``Returns:``
            |  A confirmation message on removal of datasets or samples from cohort.

        ``Errors:``
            |  ``InvalidParameterException:`` Empty or Invalid Parameters
            |  ``InvalidCohortOperationException:`` This operation is not valid as no cohort has been \
instantiated.

        |  Example-
        .. code::


                cohort.remove_from_cohort(list_of_datasets)
        """
        self.core_obj._cohort_details = self._cohort_details
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        if "entity_type" not in self._cohort_details:
            raise EmptyCohortException
        entity_type = self._cohort_details.get("entity_type")
        if entity_type == "dataset":
            self.core_obj.remove_single_from_cohort(self._cohort_path, dataset_id)
        else:
            self.core_obj.remove_multiple_from_cohort(
                self._cohort_path, dataset_id, sample_id
            )
        self._cohort_details = self.core_obj._cohort_details

    def merge_data(self, data_level: str):
        """
        Function to merge metadata (dataset, sample and feature level metadata) or data-matrix of all the samples\
or datasets in the cohort.

        Args:
            | data_level(str): identifier to specify the data to be merged - "dataset", "sample", "feature" or \
"data_matrix"
        Returns:
            | A pandas dataframe containing the merged data which is ready for analysis
        """
        self.core_obj._cohort_details = self._cohort_details
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        if not (data_level and isinstance(data_level, str)):
            raise InvalidParameterException("data_level")
        if "entity_type" not in self._cohort_details:
            raise EmptyCohortException
        if not (data_level and isinstance(data_level, str)):
            raise InvalidParameterException("data_level")
        if data_level == "sample":
            sample_df = self.merge_sample_metadata()
            return sample_df
        if data_level == "dataset":
            dataset_df = self.merge_dataset_metadata()
            return dataset_df
        if data_level == "data_matrix":
            datamatrix_df = self.merge_data_matrix()
            return datamatrix_df
        if data_level == "feature":
            feature_df = self.merge_feature_metadata()
            return feature_df
        raise InvalidCohortMergeOperation

    @deprecated(reason="use function merge_data")
    def merge_sample_metadata(self) -> pd.DataFrame:
        """
        Function to merge the sample level metadata from all the gct files in a cohort.
        Returns:
            | A pandas dataframe containing the merged metadata for analysis.

        :meta private:
        """
        self.core_obj._cohort_details = self._cohort_details
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        if "entity_type" not in self._cohort_details:
            raise EmptyCohortException
        entity_type = self._cohort_details.get("entity_type")
        if entity_type == "sample":
            return self.core_obj.merge_multiple_sample_metadata(self._cohort_path)
        else:
            return self.core_obj.merge_single_sample_metadata(self._cohort_path)

    def is_valid(self) -> bool:
        """
        This function is used to check the validity of a cohort.

        ``Returns:``
            |  A boolean result based on the validity of the cohort.

        ``Errors:``
            |  ``InvalidPathException:`` Cohort path does not represent a file or a directory.
            |  ``InvalidCohortOperationException:`` This operation is not valid as no cohort has been \
instantiated.

        .. code::


                cohort2.is_valid()
        """
        self.core_obj._cohort_details = self._cohort_details
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        return self.core_obj.is_valid(self._cohort_path)

    @deprecated(reason="use function merge_data")
    def merge_feature_metadata(self) -> pd.DataFrame:
        """
        Function to merge the feature level metadata from all the gct files in a cohort.
        Returns:
            | A pandas dataframe containing the merged metadata for analysis.

        :meta private:
        """
        self.core_obj._cohort_details = self._cohort_details
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        if "entity_type" not in self._cohort_details:
            raise EmptyCohortException
        return self.core_obj.merge_feature_metadata(self._cohort_path)

    @deprecated(reason="use function merge_data")
    def merge_dataset_metadata(self) -> pd.DataFrame:
        """
        Function to merge the dataset level metadata from all the gct files in a cohort.
        Returns:
            | A pandas dataframe containing the merged metadata for analysis.

        :meta private:
        """
        self.core_obj._cohort_details = self._cohort_details
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        if "entity_type" not in self._cohort_details:
            raise EmptyCohortException
        return self.core_obj.merge_dataset_metadata(self._cohort_path)

    @deprecated(reason="use function merge_data")
    def merge_data_matrix(self) -> pd.DataFrame:
        """
        Function to merge the data matrix metadata from all the gct files in a cohort.
        Returns:
            | A pandas dataframe containing the merged metadata for analysis.

        :meta private:
        """
        self.core_obj._cohort_details = self._cohort_details
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        if "entity_type" not in self._cohort_details:
            raise EmptyCohortException
        entity_type = self._cohort_details.get("entity_type")
        if entity_type == "sample":
            return self.core_obj.merge_multiple_data_matrix(self._cohort_path)
        else:
            return self.core_obj.merge_single_data_matrix(self._cohort_path)

    def load_cohort(self, local_path: str):
        """
        Function to load an existing cohort into an object. Once loaded, the functions described in the \
documentation can be used for the object where the cohort is loaded.

        ``Args:``
            |  ``local_path(str):`` local path of the cohort.

        ``Returns:``
            |  A confirmation message on instantiation of the cohort.

        ``Errors:``
            |  ``InvalidPathException:`` This path does not represent a file or a directory.
            |  ``InvalidCohortPathException:`` This path does not represent a Cohort.

        |  Example-
        .. code::

                cohort.load_cohort("/path/cohort_name.pco")

        """
        self.core_obj.load_cohort(local_path)
        self._cohort_path = local_path
        self._cohort_details = self.core_obj._cohort_details

    def edit_cohort(self, new_cohort_name=None, new_description=None):
        """
          This function is used to edit the cohort level metadata such as cohort name and description.

        ``Args:``
            |  ``new_cohort_name(str):`` Optional Argument: new identifier name for the cohort.
            |  ``new_description(str):`` Optional Argument: new description about the cohort.

        ``Returns:``
            |  A confirmation message on updation of cohort.

        ``Errors:``
            |  ``InvalidCohortOperationException:`` This operation is not valid as no cohort has been \
instantiated.
            |  ``CohortEditException:`` No parameter specified for editing in cohort

        |  Example-
        .. code::


                cohort.edit_cohort("edited-cohort-name","edited-cohort-description")

        """
        self.core_obj._cohort_details = self._cohort_details
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        if new_cohort_name is None and new_description is None:
            raise CohortEditException
        new_path = self.core_obj.edit_cohort(
            new_cohort_name, new_description, self._cohort_path
        )
        if new_path:
            self._cohort_path = new_path
        self._cohort_details = self.core_obj._cohort_details

    def summarize_cohort(self):
        """
        Function to return cohort level metadata and dataframe with datasets or samples added in the cohort.

        ``Returns:``
            |  A tuple with the first value as cohort metadata information (name, description and number of \
dataset(s) or sample(s) in the cohort) and the second value as dataframe containing the source, \
dataset_id or sample_id and data type available in the cohort.

        ``Errors:``
            |  ``InvalidCohortOperationException:`` This operation is not valid as no cohort has been \
instantiated.

        | Example-

        .. code::


                metadata, cohort_details = cohort.summarize_cohort()
        | metadata will contain a object like this.

        .. code::


                {
                'cohort_name': 'cohort_name',
                'number_of_samples': 6,
                'description': 'cohort_description'
                }
        | cohort detail will contain a table like that
        .. csv-table::
            :header: "", "source_omixatlas",  "datatype", "dataset_id"
            :delim: |

            0 |	tcga |	Mutation |	BRCA_Mutation_TCGA-A8-A09Q-01A-11W-A019-09
            1 |	tcga |	Mutation |	BRCA_Mutation_TCGA-AN-A0FL-01A-11W-A050-09
            2 |	tcga |	Mutation |	BRCA_Mutation_TCGA-AR-A254-01A-21D-A167-09
            3 |	tcga |	Mutation |	BRCA_Mutation_TCGA-D8-A1XO-01A-11D-A14K-09
            4 |	tcga |	Mutation |	BRCA_Mutation_TCGA-EW-A1J2-01A-21D-A13L-09
            5 |	tcga |	Mutation |	BRCA_Mutation_TCGA-LL-A50Y-01A-11D-A25Q-09
        """
        self.core_obj._cohort_details = self._cohort_details
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        return self.core_obj.summarize_cohort(self._cohort_path)

    def delete_cohort(self) -> None:
        """
        This function is used to delete a cohort.
        Returns:
            | A confirmation message on deletion of cohort
        """
        self.core_obj._cohort_details = self._cohort_details
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        self.core_obj.delete_cohort(self._cohort_path)
        self.bridge_obj._cohort_details = self.core_obj._cohort_details
        self._cohort_details = self.core_obj._cohort_details
        self._cohort_path = None

    def _assign_entity_type(self, repo_name: str, dataset_id: str) -> str:
        """
        Function to return entity type
        Input: repo_name(str) and dataset_id(str or list)
        Returns: Entity_type('dataset' for single mapped repositories,\
            'sample' for multiple mapped repositories.)
        """
        for repo, dict in self._cohort_info.items():
            if repo_name == repo:
                if dict["file_structure"] == "single":
                    cohort_entity_type = "dataset"
                elif dict["file_structure"] == "multiple":
                    cohort_entity_type = "sample"
                else:
                    cohort_entity_type = self.bridge_obj._get_entity_type(
                        repo_name, dataset_id, self._cohort_info
                    )
        return cohort_entity_type

    def create_merged_gct(self, file_path: str, file_name="") -> None:
        """
          This function is used to merge all the gct files in a cohort into a single gct file.

        ``Args:``
            |  ``file_path(str):`` the system path where the gct file is to be written.
            |  ``file_name(str):`` Optional Argument: Identifier for the merged file name, cohort name will be used by default.
        """
        self.core_obj._cohort_details = self._cohort_details
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        if "entity_type" not in self._cohort_details:
            raise EmptyCohortException
        if not isinstance(file_name, str):
            raise InvalidParameterException("file_name")
        isExists = os.path.exists(file_path)
        if not isExists:
            raise InvalidPathException
        if file_name:
            if not isinstance(file_name, str):
                raise InvalidParameterException("file_name")
            file_name = f"{file_name}.gct"
        else:
            # Fetching cohort name to be used as the file name
            cohort_info = self.summarize_cohort()
            cohort_name = cohort_info[0].get("cohort_name")
            file_name = f"{cohort_name}.gct"
        entity_type = self._cohort_details.get("entity_type")
        self.core_obj.merge_gcts(entity_type, self._cohort_path, file_name, file_path)



import json
from collections import namedtuple
from typing import Dict, Optional, List
import pandas as pd
from polly.errors import (
    InvalidSchemaJsonException,
    InvalidSyntaxForRequestException,
    EmptyPayloadException,
    RequestException,
    UnauthorizedException,
    extract_json_api_error,
)
from polly.auth import Polly
from polly import helpers, constants as const, application_error_info as app_err_info
from polly.help import example, doc
import polly.http_response_codes as http_codes
from polly.constants import SUPPORTED_ENTITY_TYPES


class Curation:
    """
    The Curation class contains wrapper functions around the models used for
    semantic annotations of string/text

    ``Args:``
        |  ``token (str):`` token copy from polly.

    .. code::


            from polly.curation import Curation
            # To use this class initialize a object like this if you are not authorised on polly
            curationObj = Curation(token)

    If you are authorised then you can initialize object without token to know about :ref:`authentication <auth>`.
    """

    Tag = namedtuple("Tag", ["name", "ontology_id", "entity_type"])
    example = classmethod(example)
    doc = classmethod(doc)

    def __init__(
        self,
        token=None,
        env="",
        default_env="polly",
    ) -> None:
        # check if COMPUTE_ENV_VARIABLE present or not
        # if COMPUTE_ENV_VARIABLE, give priority
        env = helpers.get_platform_value_from_env(
            const.COMPUTE_ENV_VARIABLE, default_env, env
        )
        self.session = Polly.get_session(token, env=env)
        self.base_url = f"https://v2.api.{self.session.env}.elucidata.io"
        self.discover_url = f"https://api.discover.{self.session.env}.elucidata.io"
        self.elastic_url = (
            f"https://api.datalake.discover.{self.session.env}.elucidata.io/elastic/v2"
        )
        self.inference_url = f"https://api.discover.{self.session.env}.elucidata.io/curations/inferences/"

    def _handle_errors(self, response):
        detail = response.get("errors")[0].get("detail", [])
        title = response.get("errors")[0].get("title", [])
        return title, detail

    def _perform_inference(
        self,
        model_name: str,
        input_data: dict,
    ) -> dict:
        """
        This is a wrapper around model inference APIs
        It serializes input_data, calls the API for the given model_name
        and returns deserialized output

        Args:
            model_name (str): one of 'normalizer', 'biobert' and 'control-perturbation'
            input_data (dict): model input

        Returns:
            dict
        """
        url = self.inference_url + model_name

        payload = {}
        payload = json.dumps({"data": {"attributes": input_data, "type": "curation"}})
        response = self.session.post(url, data=payload)
        if response.status_code != 201:
            if response.status_code == http_codes.UNAUTHORIZED:
                raise UnauthorizedException("User is unauthorized to access this")
            elif response.status_code == http_codes.BAD_REQUEST:
                title, details = extract_json_api_error(response)
                if title == app_err_info.EMPTY_PAYLOAD_CODE:
                    raise EmptyPayloadException()
                elif app_err_info.INVALID_MODEL_NAME_TITLE in title:
                    raise InvalidSyntaxForRequestException()
            elif response.status_code == http_codes.INTERNAL_SERVER_ERROR:
                raise InvalidSchemaJsonException()
            else:
                title, details = extract_json_api_error(response)
                raise Exception("Exception Occurred :" + str(details))
        try:
            response = response.json()
        except json.JSONDecodeError as e:
            raise e
        if "data" in response:
            return response.get("data")
        return response

    def standardise_entity(
        self,
        mention: str,
        entity_type: str,
        context: Optional[str] = None,
        threshold: Optional[float] = None,
    ) -> dict:
        """Map a given mention (keyword) to an ontology term.

        Args:
            mention (str): mention of an entity e.g. "Cadiac arrythmia"
            entity_type (str): Should be one of
            ['disease', 'drug', 'tissue', 'cell_type', 'cell_line', 'species', 'gene']

        Kwargs:
            context (Optional[str]): The text where the mention occurs.
            This is used to resolve abbreviations

        Returns:
            dict

        """
        data = {
            "mention": {
                "keyword": mention,
                "entity_type": entity_type,
                "threshold": threshold,
            }
        }

        if context:
            data["context"] = context
        output = self._perform_inference("normalizer", data)
        if output.get("errors", []):
            title, detail = self._handle_errors(output)
            raise RequestException(title, detail)

        if "term" not in output:
            return {
                "ontology": "CUI-less",
                "ontology_id": None,
                "name": None,
                "entity_type": entity_type,
            }

        return output.get("term", [])

    def recognise_entity(
        self,
        text: str,
        threshold: Optional[float] = None,
        normalize_output: bool = False,
    ):
        """Run an NER model on the given text. The returned value is a list of entities along with span info.

        Args:
            text (str): input text

        Kwargs:
            normalize_output (bool): whether to normalize the keywords

        Returns:
            entities (List[dict]): returns a list of spans,
            each span contains the keyword, start and end index of the keyword and the entity type
        """
        # TODO: If text is too long, break it up into chunks small enough for biobert

        payload = {"text": text}
        if threshold:
            payload["threshold"] = threshold
        response = self._perform_inference("biobert", payload)

        if "errors" in response:
            title, detail = self._handle_errors(response)
            raise RequestException(title, detail)
        try:
            entities = response.get("entities", [])
        except KeyError as e:
            raise e

        # TODO: fetch this list from the server maybe?

        if normalize_output:
            for entity in entities:

                # don't call `normalize` for unsupported entity types
                if entity.get("entity_type") not in SUPPORTED_ENTITY_TYPES:
                    entity["name"] = None
                    continue
                norm = self.standardise_entity(
                    entity["keyword"], entity["entity_type"], text
                )
                if norm.get("ontology", []) == "CUI-less":
                    entity["name"] = None
                else:
                    entity["ontology_id"] = norm["ontology"] + ":" + norm["ontology_id"]
                    entity["name"] = norm["name"]
        return entities

    def annotate_with_ontology(
        self,
        text: str,
    ) -> List[Tag]:

        """
        Tag a given piece of text. A "tag" is just an ontology term.
        Annotates with Polly supported ontologies.

        This function calls `recognise_entity` followed by `normalize`

        Args:
            text (str): Input text

        Returns:
            tags (set of tuples): set of unique tags
        """

        entities = self.recognise_entity(text, normalize_output=True)
        res = {
            self.Tag(
                e.get("name", []), e.get("ontology_id", []), e.get("entity_type", [])
            )
            for e in entities
            if e.get("name")
        }
        return list(res)

    def find_abbreviations(self, text: str) -> Dict[str, str]:
        """
        To run abbreviation detection separately.
        Internally calls _perform_inference("normalizer", data)

        Arguments:
            text -- The string to detect abbreviations in

        Returns:
            Dict with abbreviation as key and full form as value
        """
        data = {
            "mention": {"keyword": "dummykeyword", "entity_type": "gene"},
            "context": text,
        }

        response = self._perform_inference("normalizer", data)
        if "errors" in response:
            title, detail = self._handle_errors(response)
            raise RequestException(title, detail)
        try:
            output = response.get("abbreviations", [])
        except KeyError as e:
            raise e

        return output

    def assign_control_pert_labels(
        self, sample_metadata, columns_to_exclude=None
    ) -> pd.DataFrame:
        """Returns the sample metadata dataframe with 2 additional columns

        `is_control` - whether the sample is a control sample
        `control_prob` - the probability that the sample is control

        Args:
            sample_metadata (pd.DataFrame):

        Kwargs:
            columns_to_exclude (Set[str]): Any columns which don't play any
            role in determining the label,
            e.g. any arbitrary sample identifier

        Returns:
            pd.DataFrame

        """
        sample_metadata = sample_metadata.copy()

        if columns_to_exclude is None:
            columns_to_exclude = []

        cols = sample_metadata.columns.difference(columns_to_exclude)

        samples = sample_metadata[cols].to_dict("records")
        request_body = {"samples": samples}
        response = self._perform_inference("control-pertubation", request_body)
        response = {
            k: v for k, v in response.items() if k != "version" and k != "testing"
        }
        if "errors" in response:
            title, detail = self._handle_errors(response)
            raise RequestException(title, detail)
        try:
            output = pd.DataFrame(response)
        except KeyError as e:
            raise e
        sample_metadata["is_control"] = output["is_control"].values
        sample_metadata["control_prob"] = output["control_prob"].values
        return sample_metadata



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This file contains the relevant http response codes that need to be sent by
the server and various scenarios corresponding to them.
These have been compiled based on JSON:API v1.0 specifications.
For further details refer - https://jsonapi.org/format/
HTTP Response codes spec -
https://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
"""

# SUCCESS responses
"""
200 OK
Generic success status
"""
OK = 200


"""
201 CREATED
new resource created successfully but not returned in response
"""
CREATED = 201


"""
202 ACCEPTED
If an update request has been accepted for processing, but the processing has
not been completed by the time the server responds.
This can be used for async requests.
"""
ACCEPTED = 202


"""
204 NO CONTENT
* resouce created successfully with client-generated-id and
created resouce not sent in response;
* update is successful and the server doesn’t update any attributes besides
those provided and no response document;
"""
NO_CONTENT = 204


# ERROR responses
"""
400 BAD REQUEST
* server encounters a query parameter that does not follow JSON API naming
conventions, and the server does not know how to process it as a query
parameter from this specification
* request contains 'include' parameter but endpoint does not support it.
* server does not support sorting as specified in the query parameter 'sort'
* server is unable to identify a relationship path or does not support
inclusion of resources from a path
* multiple 4xx errors
"""
BAD_REQUEST = 400


"""
401 UNAUTHORIZED
Missing/invlaid JWT token
"""
UNAUTHORIZED = 401


"""
404 NOT FOUND
* processing a request to fetch a single resource that does not exist.
* processing a request to fetch a relationship link URL that does not exist.
* processing a request to modify a resource that does not exist.
* processing a request that references a related resource that does not exist.
* deletion request fails due to the resource not existing.
"""
NOT_FOUND = 404


"""
403 FORBIDDEN
* unsupported request to create a resource
* unsupported request to create a resource with a client-generated ID
* server MAY reject an attempt to do a full replacement of a
to-many relationship. In such a case, the server MUST reject the entire update,
and return a 403 Forbidden response
* unsupported request to update a resource or relationship.
* if complete replacement is not allowed by the server in PATCH request.
* If the client makes a DELETE request to a URL from a relationship link
the server MUST delete the specified members from the relationship
or return a 403 Forbidden response
"""
FORBIDDEN = 403


"""
409 CONFLICT
* processing a POST request to create a resource with a client-generated ID
that already exists.
* processing a POST request in which the resource object’s type is not among
the type(s) that constitute the collection represented by the endpoint.
* processing a PATCH request to update a resource that would
violate other server-enforced constraints
(such as a uniqueness constraint on a property other than id).
* processing a PATCH request in which the resource object’s type and id
do not match the server’s endpoint.
"""
CONFLICT = 409


"""
413 PAYLOAD TOO LARGE
the request entity is larger than limits defined by server
"""
PAYLOAD_TOO_LARGE = 413


"""
415 UNSUPPORTED MEDIA TYPE
If a request specifies the header Content-Type: application/vnd.api+json with
any media type parameters
"""
UNSUPPORTED_MEDIA_TYPE = 415


"""
405 METHOD NOT ALLOWED
method is known by the origin server but not allowed for the
requested resource
"""
METHOD_NOT_ALLOWED = 405


"""
406 NOT ACCEPTABLE
If a request’s Accept header contains the JSON:API media type and all
instances of that media type are modified with media type parameters.
"""
NOT_ACCEPTABLE = 406

"""
422 UNPROCESSABLE ENTITY
Server understands the content type of the request entity, and the
syntax of the request entity is correct, but it was unable to process
the contained instructions
"""
UNPROCESSABLE_ENTITY = 422

"""
429 TOO MANY REQUESTS
Client has sent too many requests in a given amount of time ("rate limiting").
"""
TOO_MANY_REQUESTS = 429

"""
500 INTERNAL SERVER ERROR
Generic server error
"""
INTERNAL_SERVER_ERROR = 500

"""
501 NOT IMPLEMENTED
Server does not support the functionality required to fulfill the request.
This is the appropriate response when the server does not recognize
the request method and is not capable of supporting it for any resource.
The only request methods that servers are required to support
(and therefore that must not return this code) are GET and HEAD
"""
NOT_IMPLEMENTED = 501


"""
502 Bad Gateway
Server, while acting as a gateway or proxy, received an invalid response
from the upstream server.
Note: A Gateway might refer to different things in networking and a 502 error
is usually not something you can fix, but requires a fix by the web server or
the proxies you are trying to get access through.
"""
BAD_GATEWAY = 502


"""
504 Gateway Timeout
The server, while acting as a gateway or proxy, did not receive a timely
response from an upstream server it needed to access in order to complete
the request.
"""
GATEWAY_TIMEOUT = 504



PARAM_EXCEPTION = "parameter passed is either empty or wrong datatype"
PARAM_EXCEPTION_TITLE = "parameter error"

WRONG_PARAMS_EXCEPTION = "Incorrect value of param passed"
WRONG_PARAMS_EXCEPTION_TITLE = "parameter error"

API_ERROR_EXCEPTION_TITLE = "API ERROR"
API_ERROR_EXCEPTION = ""

INCORRECT_DATA = "Data is Incorrect"

BAD_REQUEST_CODE = "bad_req"
BAD_REQUEST_TITLE = "Bad Request"
BAD_REQUEST_DETAIL = "Server could not understand the request due to invalid" " syntax"


FORBIDDEN_CODE = "forbidden"
FORBIDDEN_TITLE = "Access denied"
FORBIDDEN_DETAIL = "Access denied for requested resource"

NOT_FOUND_CODE = "resource_not_found"
NOT_FOUND_TITLE = "Resource not found"
NOT_FOUND_DETAIL = "Requested resource does not exist"

INVALID_PAYLOAD_CODE = "invalid_payload"
INVALID_PAYLOAD_TITLE = "Invalid Payload"
INVALID_PAYLOAD_DETAIL = "Parameters passed into the request are" "invalid or missing"

BAD_GATEWAY_CODE = "bad_gateway"
BAD_GATEWAY_TITLE = "Received an invalid response from the upstream service"

GATEWAY_TIMEOUT_CODE = "gateway_timeout"
GATEWAY_TIMEOUT_TITLE = "Gateway Timeout"
GATEWAY_TIMEOUT_DETAIL = "Request timed out"

EMPTY_PAYLOAD_CODE = "bad_req"
EMPTY_PAYLOAD_TITLE = "no payload found"
EMPTY_PAYLOAD_DETAIL = "Server could not understand the request due to invalid syntax"

INVALID_MODEL_NAME_TITLE = "doesn't exist"
JSON_SCHEMA_REQUEST_ERROR_DETAIL = (
    "Schema error in the json request file; update and try again"
)

REPOSITORY_LOCKED = "Repository Locked"
REPOSITORY_LOCKED_DETAIL = (
    "Your schema update is in progress. During this the repository has been locked for any data changes. "
    + "Please refer to the ingestion monitoring dashboard for visibility on this process. "
    + "Once completed, you can edit your datasets or schema. "
)



# index and level name
indexes_schema_level_map = {
    "files": "dataset",
    "gct_metadata": "sample",
    "h5ad_metadata": "sample",
}

schema_levels = ["dataset", "sample"]

schema_table_names = [
    "datasets",
    "samples",
    "features",
    "samples_singlecell",
    "features_singlecell",
    "variant_data",
]
